@@ -22,13 +22,14 @@ import threading as _threading
 from logging import DEBUG as _DEBUG
 from logging import INFO as _INFO
 from logging import getLogger
+from struct import unpack as _unpack
 
 from . import diversion as _diversion
 from . import hidpp10 as _hidpp10
 from . import hidpp20 as _hidpp20
+from . import settings_templates as _st
 from .base import DJ_MESSAGE_ID as _DJ_MESSAGE_ID
 from .common import strhex as _strhex
-from .common import unpack as _unpack
 from .i18n import _
 from .status import ALERT as _ALERT
 from .status import KEYS as _K
@@ -54,7 +55,7 @@ def process(device, notification):
     status = device.status
     assert status is not None
 
-    if device.kind is None:
+    if not device.isDevice:
         return _process_receiver_notification(device, status, notification)
 
     return _process_device_notification(device, status, notification)
@@ -161,6 +162,9 @@ def _process_device_notification(device, status, n):
     # HID++ 1.0 requests, should never get here
     assert n.sub_id & 0x80 == 0
 
+    if n.sub_id == 00:  # no-op feature notification, dispose of it quickly
+        return False
+
     # Allow the device object to handle the notification using custom
     # per-device state.
     handling_ret = device.handle_notification(n)
@@ -174,6 +178,8 @@ def _process_device_notification(device, status, n):
         else:
             return _process_hidpp10_notification(device, status, n)
 
+    # These notifications are from the device itself, so it must be active
+    device.online = True
     # At this point, we need to know the device's protocol, otherwise it's
     # possible to not know how to handle it.
     assert device.protocol is not None
@@ -187,7 +193,7 @@ def _process_device_notification(device, status, n):
         _log.warn('%s: feature notification but features not set up: %02X %s', device, n.sub_id, n)
         return False
     try:
-        feature = device.features[n.sub_id]
+        feature = device.features.get_feature(n.sub_id)
     except IndexError:
         _log.warn('%s: notification from invalid feature index %02X: %s', device, n.sub_id, n)
         return False
@@ -229,8 +235,8 @@ def _process_hidpp10_custom_notification(device, status, n):
         # message layout: 10 ix <register> <xx> <yy> <zz> <00>
         assert n.data[-1:] == b'\x00'
         data = chr(n.address).encode() + n.data
-        charge, status_text, next_charge = _hidpp10.parse_battery_status(n.sub_id, data)
-        status.set_battery_info(charge, status_text, next_charge)
+        charge, next_charge, status_text, voltage = _hidpp10.parse_battery_status(n.sub_id, data)
+        status.set_battery_info(charge, next_charge, status_text, voltage)
         return True
 
     if n.sub_id == _R.keyboard_illumination:
@@ -305,51 +311,58 @@ def _process_hidpp10_notification(device, status, n):
 
 
 def _process_feature_notification(device, status, n, feature):
+    if _log.isEnabledFor(_DEBUG):
+        _log.debug('%s: notification for feature %s, report %s, data %s', device, feature, n.sub_id >> 4, _strhex(n.data))
+
     if feature == _F.BATTERY_STATUS:
         if n.address == 0x00:
-            discharge_level = ord(n.data[:1])
-            discharge_level = None if discharge_level == 0 else discharge_level
-            discharge_next_level = ord(n.data[1:2])
-            battery_status = ord(n.data[2:3])
-            status.set_battery_info(discharge_level, _hidpp20.BATTERY_STATUS[battery_status], discharge_next_level)
+            _ignore, discharge_level, discharge_next_level, battery_status, voltage = _hidpp20.decipher_battery_status(n.data)
+            status.set_battery_info(discharge_level, discharge_next_level, battery_status, voltage)
         elif n.address == 0x10:
             if _log.isEnabledFor(_INFO):
                 _log.info('%s: spurious BATTERY status %s', device, n)
         else:
             _log.warn('%s: unknown BATTERY %s', device, n)
-        return True
 
-    if feature == _F.BATTERY_VOLTAGE:
+    elif feature == _F.BATTERY_VOLTAGE:
         if n.address == 0x00:
-            battery_level, battery_status, battery_voltage, _ignore, _ignore = _hidpp20.decipher_voltage(n.data)
-            status.set_battery_info(battery_level, battery_status, None, battery_voltage)
+            _ignore, level, nextl, battery_status, voltage = _hidpp20.decipher_battery_voltage(n.data)
+            status.set_battery_info(level, nextl, battery_status, voltage)
         else:
             _log.warn('%s: unknown VOLTAGE %s', device, n)
-        return True
 
-    if feature == _F.UNIFIED_BATTERY:
+    elif feature == _F.UNIFIED_BATTERY:
         if n.address == 0x00:
-            battery_level, battery_status, battery_voltage = _hidpp20.decipher_unified_battery(n.data)
-            status.set_battery_info(battery_level, battery_status, None, battery_voltage)
+            _ignore, level, nextl, battery_status, voltage = _hidpp20.decipher_battery_unified(n.data)
+            status.set_battery_info(level, nextl, battery_status, voltage)
         else:
             _log.warn('%s: unknown UNIFIED BATTERY %s', device, n)
-        return True
 
-    if feature == _F.SOLAR_DASHBOARD:
+    elif feature == _F.ADC_MEASUREMENT:
+        if n.address == 0x00:
+            result = _hidpp20.decipher_adc_measurement(n.data)
+            if result:
+                _ignore, level, nextl, battery_status, voltage = result
+                status.set_battery_info(level, nextl, battery_status, voltage)
+            else:  # this feature is used to signal device becoming inactive
+                status.changed(active=False)
+        else:
+            _log.warn('%s: unknown ADC MEASUREMENT %s', device, n)
+
+    elif feature == _F.SOLAR_DASHBOARD:
         if n.data[5:9] == b'GOOD':
             charge, lux, adc = _unpack('!BHH', n.data[:5])
             # guesstimate the battery voltage, emphasis on 'guess'
             # status_text = '%1.2fV' % (adc * 2.67793237653 / 0x0672)
             status_text = _hidpp20.BATTERY_STATUS.discharging
-
             if n.address == 0x00:
                 status[_K.LIGHT_LEVEL] = None
-                status.set_battery_info(charge, status_text, None)
+                status.set_battery_info(charge, None, status_text, None)
             elif n.address == 0x10:
                 status[_K.LIGHT_LEVEL] = lux
                 if lux > 200:
                     status_text = _hidpp20.BATTERY_STATUS.recharging
-                status.set_battery_info(charge, status_text, None)
+                status.set_battery_info(charge, None, status_text, None)
             elif n.address == 0x20:
                 if _log.isEnabledFor(_DEBUG):
                     _log.debug('%s: Light Check button pressed', device)
@@ -364,26 +377,19 @@ def _process_feature_notification(device, status, n, feature):
                 _log.warn('%s: unknown SOLAR CHARGE %s', device, n)
         else:
             _log.warn('%s: SOLAR CHARGE not GOOD? %s', device, n)
-        return True
 
-    if feature == _F.WIRELESS_DEVICE_STATUS:
+    elif feature == _F.WIRELESS_DEVICE_STATUS:
         if n.address == 0x00:
             if _log.isEnabledFor(_DEBUG):
                 _log.debug('wireless status: %s', n)
             reason = 'powered on' if n.data[2] == 1 else None
             if n.data[1] == 1:  # device is asking for software reconfiguration so need to change status
-                # only show a user notification if the device can change hosts
-                # as we want to notify when a device changes to this host
-                # but the only indication we get is this notification
-                alert = _ALERT.NOTIFICATION if _F.CHANGE_HOST in device.features else _ALERT.NONE
-                status.changed(active=True, alert=alert, reason=reason)
-            else:
-                _log.warn('%s: unknown WIRELESS %s', device, n)
+                alert = _ALERT.NONE
+                status.changed(active=True, alert=alert, reason=reason, push=True)
         else:
             _log.warn('%s: unknown WIRELESS %s', device, n)
-        return True
 
-    if feature == _F.TOUCHMOUSE_RAW_POINTS:
+    elif feature == _F.TOUCHMOUSE_RAW_POINTS:
         if n.address == 0x00:
             if _log.isEnabledFor(_INFO):
                 _log.info('%s: TOUCH MOUSE points %s', device, n)
@@ -408,7 +414,7 @@ def _process_feature_notification(device, status, n, feature):
         if n.address == 0x00:
             if _log.isEnabledFor(_DEBUG):
                 cid1, cid2, cid3, cid4 = _unpack('!HHHH', n.data[:8])
-                _log.debug('%s: diverted controls pressed: %i, %i, %i, %i', device, cid1, cid2, cid3, cid4)
+                _log.debug('%s: diverted controls pressed: 0x%x, 0x%x, 0x%x, 0x%x', device, cid1, cid2, cid3, cid4)
         elif n.address == 0x10:
             if _log.isEnabledFor(_DEBUG):
                 dx, dy = _unpack('!hh', n.data[:4])
@@ -427,15 +433,16 @@ def _process_feature_notification(device, status, n, feature):
                 periods = flags & 0x0f
                 _log.info('%s: WHEEL: res: %d periods: %d delta V:%-3d', device, high_res, periods, delta_v)
         elif (n.address == 0x10):
+            ratchet = ord(n.data[:1]) & 0x01
             if _log.isEnabledFor(_INFO):
-                flags = ord(n.data[:1])
-                ratchet = flags & 0x01
                 _log.info('%s: WHEEL: ratchet: %d', device, ratchet)
+            from solaar.ui.config_panel import change_setting  # prevent circular import
+            setting = next((s for s in device.settings if s.name == _st.ScrollRatchet.name), None)
+            if setting:
+                change_setting(device, setting, [2 if ratchet else 1])
         else:
             if _log.isEnabledFor(_INFO):
                 _log.info('%s: unknown WHEEL %s', device, n)
 
     _diversion.process_notification(device, status, n, feature)
-
-    if _log.isEnabledFor(_DEBUG):
-        _log.debug('%s: notification for feature %r, report %s, data %s', device, feature, n.sub_id >> 4, _strhex(n.data))
+    return True
